{
    Файл является частью библиотеки RubiRobot

    © 2017-2018 Слинкин Д.А.

    Обработка двигателей контроллера Lego EV3

    Подробная информация о лицензировании находится в файле README,
    входящем в состав библиотеки.

    ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
    КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
    ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
    КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ 
    ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
    ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
    В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
    ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
    ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

 **********************************************************************}


{$SMARTLINK ON}
{$mode objfpc}
{$H+}

unit rubiroMotors;


interface
uses BaseUnix,uev3Devices,uev3sysfs,uev3;
{

    run-forever: Causes the motor to run until another command is sent.
    run-to-abs-pos: Runs the motor to an absolute position specified by``position_sp`` and then stops the motor using the command specified in stop_action.
    run-to-rel-pos: Runs the motor to a position relative to the current position value. The new position will be current position + position_sp. When the new position is reached, the motor will stop using the command specified by stop_action.
    run-timed: Run the motor for the amount of time specified in time_sp and then stops the motor using the command specified by stop_action.
    run-direct: Runs the motor using the duty cycle specified by duty_cycle_sp. Unlike other run commands, changing duty_cycle_sp while running will take effect immediately.
    stop: Stop any of the run commands before they are complete using the command specified by stop_action.
    reset: Resets all of the motor parameter attributes to their default values. This will also have the effect of stopping the motor.
}

{
    running: Power is being sent to the motor.
    ramping: The motor is ramping up or down and has not yet reached a constant output level.
    holding: The motor is not turning, but rather attempting to hold a fixed position.
    overloaded: The motor is turning as fast as possible, but cannot reach its speed_sp.
    stalled: The motor is trying to run but is not turning at all.
}

type  TEv3mCommand=(cRunForever,cRunToAbsPos,cRunToRelPos,cRunTimed,cRunDirect,cStop,cReset);
const ev3mCommands:array[TEv3mCommand] of string[20]=('run-forever','run-to-abs-pos','run-to-rel-pos','run-timed','run-direct','stop','reset');

type  TEv3mStopAction=(saCoast,saBrake,saHold);
const ev3mStopActions:array[TEv3mStopAction] of string[20]=('coast', 'brake', 'hold');

type  TEv3mPolarity=(pNormal,pInversed);
const ev3mPolarities:array[TEv3mPolarity] of string[20]=('normal', 'inversed');

type  TEv3mState=(sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled);
      TEv3mStates=set of TEv3mState;
const ev3mStates:array[TEv3mState]of string[20]=('','running','ramping','holding','overloaded','stalled');



type

  { TEv3Motor }

  TEv3Motor=class(TEv3OneDeviceHelper)
   private
    procedure resetWriteInterval;
   protected
    procedure afterConstruction;override;
   public
   procedure setIntervals(R,W:integer);override;
  end;

  { TEv3TachoMotor }
  TEv3TachoMotorClass=class of TEv3TachoMotor;
  TEv3TachoMotor=class(TEv3Motor)
    private
     fmaxSpeed:cint;
     _fspeed:cint;
     fstopaction:TEv3mStopAction;
     fStopChanged:boolean;
     fpolarity:TEv3mPolarity;
     fcount_per_rot:cint;
     function getcount_per_rot: integer;
     function getMaxSpeed: cint;
     procedure setFSpeed(AValue: cint);
    protected
     procedure deviceRefreshed(dev:TEv3Device);override;
     property maxSpeed:cint read getMaxSpeed;
     procedure reversePolarity;
     property fspeed:cint read _fspeed write setFSpeed;
     {
       Устанавливает реверс
       Действует при запуске мотора.
       На текущее вращение мотора влияния не оказывает.
     }
     function setReverse(yes:variant):TEv3TachoMotor;
    public
     class function varToSA(sa:variant):TEv3mStopAction;
     class function varToState(s:variant):TEv3mStates;
   //  function wait()
     // ждет события из списка событий: sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled
     function waitEvent(event:ansistring; timeout:dword=high(dword)):boolean; override;
    public
     procedure setIntervals(R,W:integer);override;

     property count_per_rot:integer read getcount_per_rot;
     function wait():TEv3TachoMotor; // ждет остановки мотора

     {
      Устанавливает способ остановки мотора.
      Действует при запуске мотора.
      На текущее вращение мотора влияния не оказывает.
      Варианты:
      1) Остановка без отключения подачи электроэнергии,
         средний тормозной путь
         saBrake,'brake','тормоз','тормози','тормозить',1
      2) Остановка с удержанием позиции, минимальный тормозной путь
         saHold,'hold','стоп','стоять','стой',2
      0) Движение накатом, с отключением подачи электроэнергии,
         максимальный тормозной путь
         saCoast,0,'накат' и другие варианты, не входящие в 1) и 2)
     }
     function setHowStop(sa:variant):TEv3TachoMotor;
     {
      Возвращает способ остановки мотора
       1 - Остановка без отключения подачи электроэнергии,
           средний тормозной путь
       2 - Остановка с удержанием позиции, минимальный тормозной путь
       0 - Движение накатом, с отключением подачи электроэнергии,
           максимальный тормозной путь
     }
     function howStop(out sa:variant):TEv3TachoMotor;
     function howStop:variant;

     {
       Устанавливает и возвращает текущую позицию мотора
     }
     function setPosition(p:variant):TEv3TachoMotor;
     function position(out p:variant):TEv3TachoMotor;
     function position:variant;

     {
       Устанавливает скорость вращения от -100% до 100%
       Действует при запуске мотора.
       На текущее вращение мотора влияния не оказывает.
     }
     function setSpeed(sp:variant):TEv3TachoMotor;
     {
       Возвращает установленную ранее скорость вращения от -100 до 100
     }
     function Speed(out sp:variant):TEv3TachoMotor;
     function Speed:variant;

     {
       Возвращает реальную скорость вращения от 0 до 100
     }
     function currentSpeed(out sp:variant):TEv3TachoMotor;
     function currentSpeed:variant;

     {
       Возвращает 1, если установлен реверс и 0 в противном случае
     }
     function Reverse(out yes:variant):TEv3TachoMotor;
     function Reverse:variant;

     {
       Останавливает мотор, способ остановки - см. setHowStop
     }
     function Stop(sa:variant):TEv3TachoMotor;
     function Stop():TEv3TachoMotor;

     {
       Запускает мотор (aсинхронный вызов). Возможные параметры - скорость и способ остановки
       по умолчанию при вызове stop()
     }
     function Run(sp:variant; sa:variant):TEv3TachoMotor;
     function Run(sp:variant):TEv3TachoMotor;
     function Run():TEv3TachoMotor;

     {
       Запускает мотор вплоть до достижения абсолютной позиции(aсинхронный вызов).
       Возможные параметры - скорость и способ остановки

     }
     function RunToAbsPos(pos:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunToAbsPos(pos:variant; sp:variant):TEv3TachoMotor;
     function RunToAbsPos(pos:variant):TEv3TachoMotor;
    {
     function RunToRelPos(pos:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunToRelPos(pos:variant; sp:variant):TEv3TachoMotor;
     function RunToRelPos(pos:variant):TEv3TachoMotor;
     }
     {
       Запускает мотор на количество миллисекунд (aсинхронный вызов).
       Возможные параметры - скорость и способ остановки
     }
     function RunTime(ms:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunTime(ms:variant; sp:variant):TEv3TachoMotor;
     function RunTime(ms:variant):TEv3TachoMotor;



     {
       Запускает мотор на количество градусов (aсинхронный вызов).
       Возможные параметры - скорость и способ остановки
     }
     function RunDeg(deg:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunDeg(deg:variant; sp:variant):TEv3TachoMotor;
     function RunDeg(deg:variant):TEv3TachoMotor;

     {
       Запускает мотор на количество оборотов (aсинхронный вызов).
       Возможные параметры - скорость и способ остановки
     }
     function RunTurn(turn:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunTurn(turn:variant; sp:variant):TEv3TachoMotor;
     function RunTurn(turn:variant):TEv3TachoMotor;

  end;

  { TEv3MediumMotor }

  TEv3MediumMotor=class(TEv3TachoMotor)
   protected
    class function deviceType:TEv3DeviceType;override;
  end;


  { TEv3LargeMotor }

  TEv3LargeMotor=class(TEv3TachoMotor)
  protected
   class function deviceType:TEv3DeviceType;override;
  end;

  { TEv3Rule }

  TEv3Rule=class(TEv3)
   private
    _fspeed:cint;
    fdirect:cint;
    procedure setFSpeed(AValue: cint);
   protected
    property fspeed:cint read _fspeed write setFSpeed;
    // устанавливает реверс (см.TEv3TachoMotor)
    function setReverse(yes:variant):TEv3Rule;
   public
    L,R:TEv3TachoMotor;
   public
    function Connected:boolean;
    procedure setIntervals(RR,WW:integer);

    // ждет остановки моторов
    function wait():TEv3Rule;

    constructor create(LPort,RPort:variant; MotorClass:TEv3TachoMotorClass);
    constructor create(LPort,RPort:variant);
    constructor create(MotorClass:TEv3TachoMotorClass);
    constructor create();
    destructor destroy;override;


    // Устанавливает способ остановки моторов (см.TEv3TachoMotor)
    function setHowStop(sa:variant):TEv3Rule;
    // Возвращает способ остановки моторов (см.TEv3TachoMotor)
    function howStop(out sa:variant):TEv3Rule;
    function howStop:variant;

    {
      Устанавливает скорость вращения от -100% до 100% (см.TEv3TachoMotor)
    }
    function setSpeed(sp:variant):TEv3Rule;
    {
      Возвращает установленную ранее скорость вращения от -100 до 100
      (см.TEv3TachoMotor)
    }
    function Speed(out sp:variant):TEv3Rule;
    function Speed:variant;

    {
      Возвращает установленную ранее скорость вращения для левого колеса
    }
    function SpeedLeft(out sp:variant):TEv3Rule;
    function SpeedLeft:variant;

    {
      Возвращает установленную ранее скорость вращения для правого колеса
    }
    function SpeedRight(out sp:variant):TEv3Rule;
    function SpeedRight:variant;

    // возвращает реверс (см.TEv3TachoMotor)
    function Reverse(out yes:variant):TEv3Rule;
    function Reverse:variant;

    // останавливает оба мотора (см.TEv3TachoMotor)
    function Stop(sa:variant):TEv3Rule;
    function Stop():TEv3Rule;

    {
     запускает оба мотора (aсинхронный вызов) (см.TEv3TachoMotor)
     direct - направление движения
     Примеры поведения рулевого управления при положительной скорости
     и значения direct:
      100 - поворот вправо вокруг оси,
      -100 - поворот влево вокруг оси,
      50 - поворот вправо вокруг правого колеса,
      -50 - поворот влево вокруг левого колеса
    }
    function Run(direct:variant; sp:variant; sa:variant):TEv3Rule;
    function Run(direct:variant; sp:variant):TEv3Rule;
    function Run(direct:variant):TEv3Rule;

    {
      Запускает моторы на количество миллисекунд (aсинхронный вызов).
      Возможные параметры - скорость и способ остановки
    }
    function RunTime(direct:variant; ms:variant; sp:variant; sa:variant):TEv3Rule;
    function RunTime(direct:variant; ms:variant; sp:variant):TEv3Rule;
    function RunTime(direct:variant; ms:variant):TEv3Rule;

    {
      Запускает моторы на количество градусов (БЛОКИРУЮЩИЙ вызов).
      Возможные параметры - скорость и способ остановки
    }
    function RunDegWait(direct:variant; deg:variant; sp:variant; sa:variant):TEv3Rule;
    function RunDegWait(direct:variant; deg:variant; sp:variant):TEv3Rule;
    function RunDegWait(direct:variant; deg:variant):TEv3Rule;

    {
      Запускает моторы на количество оборотов (БЛОКИРУЮЩИЙ вызов).
      Возможные параметры - скорость и способ остановки
    }
    function RunTurnWait(direct:variant; turn:variant; sp:variant; sa:variant):TEv3Rule;
    function RunTurnWait(direct:variant; turn:variant; sp:variant):TEv3Rule;
    function RunTurnWait(direct:variant; turn:variant):TEv3Rule;
  end;

 {
  номерные функции возвращают двигатель по порядку подключения вне зависимости
  от номера порта. Либо nil.
 }

 {
  функции с параметром возвращают двигатель по номеру порта.
  идентификация портов двигателей:
  1(5)-4(8),'A'-'D', 'outA'-'outD' - вне зависимости от регистра.
 }

  function ev3Motor1:TEv3LargeMotor;
  function ev3Motor2:TEv3LargeMotor;
  function ev3Motor3:TEv3LargeMotor;
  function ev3Motor4:TEv3LargeMotor;
  function ev3Motor(index:integer=-1):TEv3LargeMotor;
  function ev3Motor(index:string):TEv3LargeMotor;

  function ev3MMotor1:TEv3MediumMotor;
  function ev3MMotor2:TEv3MediumMotor;
  function ev3MMotor3:TEv3MediumMotor;
  function ev3MMotor4:TEv3MediumMotor;
  function ev3MMotor(index:integer=-1):TEv3MediumMotor;
  function ev3MMotor(index:string):TEv3MediumMotor;

  function ev3Rule(index1:integer=-1; index2:integer=-1):TEv3Rule;
  function ev3Rule(index1,index2:string):TEv3Rule;

  function ev3MRule(index1:integer=-1; index2:integer=-1):TEv3Rule;
  function ev3MRule(index1,index2:string):TEv3Rule;

implementation
uses
  SysUtils,Unix,uev3utf8,variants,dateutils;

function ev3Motor1: TEv3LargeMotor;
begin
   result:=ev3Motor(-1);
end;

function ev3Motor2: TEv3LargeMotor;
begin
 result:=ev3Motor(-2);
end;

function ev3Motor3: TEv3LargeMotor;
begin
 result:=ev3Motor(-3);
end;

function ev3Motor4: TEv3LargeMotor;
begin
 result:=ev3Motor(-4);
end;

function ev3Motor(index: integer): TEv3LargeMotor;
begin
 if (index<0) or (index>4) then result:=ev3Helper(index,TEv3LargeMotor) as TEv3LargeMotor
 else result:=ev3Helper(index+4,TEv3LargeMotor) as TEv3LargeMotor;
end;

function ev3Motor(index: string): TEv3LargeMotor;
begin
  result:=ev3Motor(ord(TEv3DeviceHelper.varToPort(index)));
end;

//-------
function ev3MMotor1: TEv3MediumMotor;
begin
   result:=ev3MMotor(-1);
end;

function ev3MMotor2: TEv3MediumMotor;
begin
 result:=ev3MMotor(-2);
end;

function ev3MMotor3: TEv3MediumMotor;
begin
 result:=ev3MMotor(-3);
end;

function ev3MMotor4: TEv3MediumMotor;
begin
 result:=ev3MMotor(-4);
end;

function ev3MMotor(index: integer): TEv3MediumMotor;
begin
 if (index<0) or (index>4) then result:=ev3Helper(index,TEv3MediumMotor) as TEv3MediumMotor
 else result:=ev3Helper(index+4,TEv3MediumMotor) as TEv3MediumMotor;
end;

function ev3MMotor(index: string): TEv3MediumMotor;
begin
 result:=ev3MMotor(ord(TEv3DeviceHelper.varToPort(index)));
end;
// -----------------

var
ev3FirstRule:TEv3Rule=nil;
ev3Rules:array[outA..outD,outA..outD]of Tev3Rule=(
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil)
);

function ev3Rule(index1: integer; index2: integer): TEv3Rule;
 var lp,rp:TEv3Port;
begin
  if index1*index2<0 then exit(nil); // неодинаковый способ инициализации моторов запрещен
  if index1<0 then begin
   if assigned(ev3FirstRule) then exit(ev3FirstRule);
   lp:=ev3Motor1.port;
   if not (lp in [outA..outD]) then exit(nil);
   rp:=ev3Motor2.port;
   if not (rp in [outA..outD]) then exit(nil);
   ev3FirstRule:=ev3rule(ord(lp),ord(rp));
   exit(ev3FirstRule);
  end;
  // теперь - по переданным портам
  if index1<5 then inc(index1,4);
  if index2<5 then inc(index2,4);
  lp:=TEv3Port(index1);
  rp:=TEv3Port(index2);
  if assigned(ev3Rules[lp,rp]) then exit(ev3Rules[lp,rp]);
  ev3Rules[lp,rp]:=TEv3Rule.Create(lp,rp);
  result:=ev3Rules[lp,rp];
end;

function ev3Rule(index1, index2: string): TEv3Rule;
begin
 result:=ev3Rule(ord(TEv3DeviceHelper.varToPort(index1)),ord(TEv3DeviceHelper.varToPort(index2)));
end;

//-----------------
var
ev3MFirstRule:TEv3Rule=nil;
ev3MRules:array[outA..outD,outA..outD]of Tev3Rule=(
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil)
);

function ev3MRule(index1: integer; index2: integer): TEv3Rule;
var lp,rp:TEv3Port;
begin
 if index1*index2<0 then exit(nil); // неодинаковый способ инициализации моторов запрещен
 if index1<0 then begin
  if assigned(ev3FirstRule) then exit(ev3FirstRule);
  lp:=ev3MMotor1.port;
  if not (lp in [outA..outD]) then exit(nil);
  rp:=ev3MMotor2.port;
  if not (rp in [outA..outD]) then exit(nil);
  ev3MFirstRule:=ev3Mrule(ord(lp),ord(rp));
  exit(ev3MFirstRule);
 end;
 // теперь - по переданным портам
 if index1<5 then inc(index1,4);
 if index2<5 then inc(index2,4);
 lp:=TEv3Port(index1);
 rp:=TEv3Port(index2);
 if assigned(ev3MRules[lp,rp]) then exit(ev3MRules[lp,rp]);
 ev3MRules[lp,rp]:=TEv3Rule.Create(lp,rp,TEv3MediumMotor);
 result:=ev3MRules[lp,rp];
end;

function ev3MRule(index1, index2: string): TEv3Rule;
begin
 result:=ev3MRule(ord(TEv3DeviceHelper.varToPort(index1)),ord(TEv3DeviceHelper.varToPort(index2)));
end;

{ TEv3LargeMotor }

class function TEv3LargeMotor.deviceType: TEv3DeviceType;
begin
 Result:=ev3tLTachoMotor;
end;


{ TEv3Motor }

procedure TEv3Motor.resetWriteInterval;
 var i:TEv3DeviceAttribute;
begin
  for i in TEv3DeviceAttribute do
   if assigned(ev.Attrs[i]) then ev.Attrs[i].skipInterval:=true;
end;

procedure TEv3Motor.afterConstruction;
begin
  resetWriteInterval;
end;

procedure TEv3Motor.setIntervals(R, W: integer);
begin
  // -----
end;


{ TEv3Rule }

constructor TEv3Rule.create(LPort, RPort: variant;
  MotorClass: TEv3TachoMotorClass);
begin
  inherited create;
  L:=ev3Helper(ord(TEv3DeviceHelper.varToPort(lport)),MotorClass) as TEv3TachoMotor;
  if not assigned(L) then fail;
  R:=ev3Helper(ord(TEv3DeviceHelper.varToPort(rport)),MotorClass) as TEv3TachoMotor;
  if not assigned(R) then fail;
end;

constructor TEv3Rule.create(LPort, RPort: variant);
begin
 create(LPort, RPort, TEv3LargeMotor);
end;

constructor TEv3Rule.create(MotorClass: TEv3TachoMotorClass);
begin
  inherited create;
  L:=ev3Helper(-1,MotorClass) as TEv3TachoMotor;
  if not assigned(L) then fail;
  R:=ev3Helper(-2,MotorClass) as TEv3TachoMotor;
  if not assigned(R) then begin L.free; fail; end;
end;

constructor TEv3Rule.create;
begin
 create(TEv3LargeMotor);
end;


destructor TEv3Rule.destroy;
begin
  inherited destroy;
end;

function TEv3Rule.setHowStop(sa: variant): TEv3Rule;
begin
  result:=self;
  L.setHowStop(sa); R.setHowStop(sa);
end;

function TEv3Rule.howStop(out sa: variant): TEv3Rule;
begin
 result:=self;
 L.HowStop(sa);
end;

function TEv3Rule.howStop: variant;
begin
 result:=L.howStop;
end;

function TEv3Rule.setSpeed(sp: variant): TEv3Rule;
begin
 result:=self;
 fspeed:=sp;
 if fdirect<-100 then fdirect:=-100;
 if fdirect>100 then fdirect:=100;
 if fdirect=0 then begin
   L.setSpeed(sp); R.setSpeed(sp);
 end else
 if fdirect<0 then begin
   L.setSpeed(sp+(fdirect*double(sp)/50));
   R.setSpeed(sp);
 end
 else begin
   L.setSpeed(sp);
   R.setSpeed(sp+(-fdirect*double(sp)/50));
 end;
end;

function TEv3Rule.Speed(out sp: variant): TEv3Rule;
begin
 result:=self;
 sp:=self.Speed;
end;

function TEv3Rule.Speed: variant;
begin
  result:=fspeed;
end;

function TEv3Rule.SpeedLeft(out sp: variant): TEv3Rule;
begin
 result:=self;
 L.Speed(sp);
end;

function TEv3Rule.SpeedLeft: variant;
begin
 result:=L.speed;
end;

function TEv3Rule.SpeedRight(out sp: variant): TEv3Rule;
begin
 result:=self;
 R.Speed(sp);
end;

function TEv3Rule.SpeedRight: variant;
begin
 result:=R.speed;
end;


function TEv3Rule.setReverse(yes: variant): TEv3Rule;
begin
 result:=self;
 L.setReverse(yes);
 R.setReverse(yes);
end;

function TEv3Rule.Reverse(out yes: variant): TEv3Rule;
begin
 result:=self;
 L.Reverse(yes);
end;

function TEv3Rule.Reverse: variant;
begin
 result:=L.Reverse;
end;

procedure TEv3Rule.setFSpeed(AValue: cint);
begin
  if avalue>100 then avalue:=100
  else
  if avalue<-100 then avalue:=-100;
  if _fspeed=AValue then Exit;
  _fspeed:=AValue;
end;

function TEv3Rule.Connected: boolean;
begin
  if not assigned(self) then exit(false);
  result:=l.Connected and r.connected;
end;

procedure TEv3Rule.setIntervals(RR, WW: integer);
begin
  L.setIntervals(RR,WW);
  R.setIntervals(RR,WW);
end;

function TEv3Rule.wait: TEv3Rule;
begin
  result:=self;
  if fdirect>0 then begin
   R.Wait;
   L.Stop;
  end
  else begin
   L.Wait;
   R.Stop;
  end;
  //waitAnyEvent([self,sStoped,self,sHolding,self,sStalled]);
  //sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled
end;


function TEv3Rule.Stop(sa: variant): TEv3Rule;
begin
 result:=self;
 L.Stop(sa);
 R.Stop(sa);
end;

function TEv3Rule.Stop: TEv3Rule;
begin
 result:=self;
 L.Stop();
 R.Stop();
end;

function TEv3Rule.Run(direct: variant; sp: variant; sa: variant): TEv3Rule;
begin
  setHowStop(sa);
  result:=self.run(direct,sp);
end;

function TEv3Rule.Run(direct: variant; sp: variant): TEv3Rule;
begin
 result:=self;
 fdirect:=direct;
 setSpeed(sp);
 L.Run;
 R.Run;
end;

function TEv3Rule.Run(direct: variant): TEv3Rule;
begin
 result:=self.Run(direct,speed);
end;

function TEv3Rule.RunTime(direct: variant; ms: variant; sp: variant; sa: variant
  ): TEv3Rule;
begin
 setHowStop(sa);
 result:=self.RunTime(direct,ms,sp);
end;

function TEv3Rule.RunTime(direct: variant; ms: variant; sp: variant): TEv3Rule;
begin
 result:=self;
 fdirect:=direct;
 setSpeed(sp);
 L.RunTime(ms);
 R.RunTime(ms);
end;

function TEv3Rule.RunTime(direct: variant; ms: variant): TEv3Rule;
begin
 result:=self.RunTime(direct,ms,speed);
end;

function TEv3Rule.RunDegWait(direct: variant; deg: variant; sp: variant; sa: variant
  ): TEv3Rule;
begin
 setHowStop(sa);
 result:=self.RunDegWait(direct,deg,sp);
end;

function TEv3Rule.RunDegWait(direct: variant; deg: variant; sp: variant): TEv3Rule;
begin
 result:=RunTurnWait(direct,deg/360.0,sp);
end;

function TEv3Rule.RunDegWait(direct: variant; deg: variant): TEv3Rule;
begin
 result:=self.RunDegWait(direct,deg,speed);
end;

function TEv3Rule.RunTurnWait(direct: variant; turn: variant; sp: variant;
  sa: variant): TEv3Rule;
begin
 setHowStop(sa);
 result:=self.RunTurnWait(direct,turn,sp);
end;

function TEv3Rule.RunTurnWait(direct: variant; turn: variant; sp: variant
  ): TEv3Rule;
begin
 result:=self;
 fdirect:=direct;
 setSpeed(sp);
 if fdirect=0 then begin
  if turn<0 then R.setFSpeed(-R.fspeed);
  R.RunTurn(turn);
  L.RunTurn(turn).wait();
 end else
 if fdirect>=0 then begin
//  writeln(r.fspeed,' ',l.fspeed);
  if turn<0 then R.setFSpeed(-R.fspeed);
  R.Run;
//  R.RunTurn(turn);
  L.RunTurn(turn).wait();
 // writeln('hello');
  R.Stop;
 end else begin
  //if turn>=0 then L.RunTurn(high(smallint)) else L.RunTurn(low(smallint)) ;
  if turn<0 then L.setFSpeed(-L.fspeed);
  L.Run;
  R.RunTurn(turn).wait();
  L.Stop;
 end;

end;

function TEv3Rule.RunTurnWait(direct: variant; turn: variant): TEv3Rule;
begin
 result:=self;
 self.RunTurnWait(direct,Turn,speed);
end;

{ TEv3MediumMotor }

class function TEv3MediumMotor.deviceType: TEv3DeviceType;
begin
  Result:=ev3tMTachoMotor;
end;



{ TEv3TachoMotor }

function TEv3TachoMotor.getMaxSpeed: cint;
begin
  if fmaxspeed=0 then fmaxspeed:=ev.Attrs[ev3amax_speed].iData;
  result:=fmaxspeed;
end;

function TEv3TachoMotor.getcount_per_rot: integer;
begin
  if fcount_per_rot=0 then fcount_per_rot:=ev.Attrs[ev3acount_per_rot].iData;
  exit(fcount_per_rot);
end;

procedure TEv3TachoMotor.setFSpeed(AValue: cint);
begin
  if avalue>100 then avalue:=100
  else
  if avalue<-100 then avalue:=-100;
  if _fspeed=AValue then Exit;
  _fspeed:=AValue;
end;

procedure TEv3TachoMotor.deviceRefreshed(dev: TEv3Device);
begin
  inherited deviceRefreshed(dev);
  dev.Attrs[ev3acommand].sData:=ev3mCommands[cReset];
  dev.Attrs[ev3astop_action].sData:=ev3mStopActions[fstopaction];
  dev.Attrs[ev3apolarity].sData:=ev3mPolarities[fpolarity];
end;

procedure TEv3TachoMotor.reversePolarity;
begin
  if reverse then setReverse(0) else setReverse(1);
end;


class function TEv3TachoMotor.varToSA(sa: variant): TEv3mStopAction;
begin
 case utf8lowercase(trim(varToStr(sa))) of
  'brake','тормоз','тормози','тормозить','1':result:=saBrake;
  'hold','стоп','стоять','стой','2':result:=saHold
  else result:=saCoast;
 end;
end;


procedure outSet(s:TEv3mStates);
 var i:TEv3mState;
begin
 for i in s do write(i,' ');
 writeln;
end;

class function TEv3TachoMotor.varToState(s: variant): TEv3mStates;
 var us:string;
     st:TEv3mState;
begin
 us:=utf8lowercase(trim(varToStr(s)));
 result:=[];
 // sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled
 for st in TEv3mState do begin
  // writeln(us,' == ',ev3mStates[st],' +- ', pos(ev3mStates[st],us)<>0 );
   if (pos(ev3mStates[st],us)<>0)
   //(ev3mStates[st]=us)
   or(us=inttostr(ord(st))) then include(result,st);
 end;
 if result<>[] then exit;
 case us of
  'работа':include(result,sRunning);
  'разгон':include(result,sRamping);
  'стоп':include(result,sHolding);
  'перегрузка':include(result,sOverloaded);
  'блок':include(result,sStalled)
  else include(result,sStoped);
 end;
end;


function TEv3TachoMotor.wait: TEv3TachoMotor;
begin
  result:=self;
//  sleep(100);
 //writeln('beg');
  waitAnyEvent([self,sStoped,self,sHolding,self,sStalled]);
 // writeln('end');
//  waitAnyEvent([self,sStoped]);//,self,sHolding,self,sStalled]);
  //sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled
end;



function TEv3TachoMotor.waitEvent(event: ansistring; timeout: dword): boolean;
var st:TEv3mStates;
    s:string;
begin
 st:=varToState(event);
 timerStart();
 while true do begin
  s:=ev.Attrs[ev3astate].sData;
//  writeln('+++ ',s);
//  write('++++ '); outset(varToState(s));
//  write('+++++ '); outset(st);
  if varToState(s)*st<>[] then begin sleep(10);//write('!!!!! ',ev.Attrs[ev3aposition_sp].sData); outSet(varToState(s)*st);
    exit(true);end;
  //if MilliSecondsBetween(now(),dt)>=timeout then exit(false);
  // v0.2
  if timer()>timeout then exit(false);
  sleep(waitInterval);
 end;
end;

procedure TEv3TachoMotor.setIntervals(R, W: integer);
begin
 ev.Attrs[ev3astop_action].setIntervals(R,W);
 ev.Attrs[ev3aposition].setIntervals(R,W);
 ev.Attrs[ev3aspeed_sp].setIntervals(R,W);
 ev.Attrs[ev3aspeed].setIntervals(R,W);
 ev.Attrs[ev3apolarity].setIntervals(R,W);
 ev.Attrs[ev3acommand].setIntervals(R,W);
 ev.Attrs[ev3atime_sp].setIntervals(R,W);
 ev.Attrs[ev3aposition_sp].setIntervals(R,W);
end;

function TEv3TachoMotor.setHowStop(sa: variant): TEv3TachoMotor;
begin
  result:=self;
  fstopaction:=varToSA(sa);
  ev.Attrs[ev3astop_action].sData:=ev3mStopActions[fStopAction];
end;

function TEv3TachoMotor.howStop(out sa: variant): TEv3TachoMotor;
begin
 result:=self;
 sa:=self.howStop;
end;

function TEv3TachoMotor.howStop: variant;
begin
 result:=varToSA(ev.Attrs[ev3astop_action].sData);
end;

function TEv3TachoMotor.setPosition(p: variant): TEv3TachoMotor;
begin
 result:=self;
 ev.Attrs[ev3aposition].sData:=p;
end;

function TEv3TachoMotor.position(out p: variant): TEv3TachoMotor;
begin
 result:=self;
 p:=ev.Attrs[ev3aposition].iData;
end;

function TEv3TachoMotor.position: variant;
begin
 result:=ev.Attrs[ev3aposition].iData;
end;

function TEv3TachoMotor.setSpeed(sp: variant): TEv3TachoMotor;
begin
 result:=self;
 fspeed:=sp;
 if sp<0 then setReverse(1) else setReverse(0);
 ev.Attrs[ev3aspeed_sp].iData:=(fspeed*maxSpeed) div 100;
end;

function TEv3TachoMotor.Speed(out sp: variant): TEv3TachoMotor;
begin
  result:=self;
  sp:=self.speed;
end;

function TEv3TachoMotor.Speed: variant;
begin
 result:=fspeed;
end;

function TEv3TachoMotor.currentSpeed(out sp: variant): TEv3TachoMotor;
begin
 result:=self;
 sp:=self.currentSpeed;
end;

function TEv3TachoMotor.currentSpeed: variant;
begin
 result:=(ev.Attrs[ev3aspeed].iData*100) div maxSpeed;
end;

function TEv3TachoMotor.setReverse(yes: variant): TEv3TachoMotor;
begin
 result:=self;
// writeln('Reverse = ',TEv3mPolarity(varToYes(yes)));
 ev.Attrs[ev3apolarity].sData:=ev3mPolarities[TEv3mPolarity(varToYes(yes))];
end;

function TEv3TachoMotor.Reverse(out yes: variant): TEv3TachoMotor;
begin
 result:=self;
 yes:=self.reverse;
end;

function TEv3TachoMotor.Reverse: variant;
begin
 result:=ev3mPolarities[pInversed]=ev.Attrs[ev3apolarity].sData;
end;

function TEv3TachoMotor.Stop(sa: variant): TEv3TachoMotor;
 var tmp:TEv3mStopAction;
begin
 result:=self;
 tmp:=varToSA(sa);
 if tmp<>fstopaction then fStopChanged:=true;
 setHowStop(sa);
 self.stop;
end;

function TEv3TachoMotor.Stop: TEv3TachoMotor;
 var s:string;
     st:TEv3mStates;
begin
  result:=self;
  //writeln(fstopaction);

  st:=varToState(ev.Attrs[ev3astate].sData);
  //outset(st);
  if sStoped in st then exit;
  // повторная остановка с удержанием, когда мотор уже стоит с удержанием,
  // даст небольшой сдвиг вперед,
  // поэтому блокируем сие безосмысленно-вредоносное действо
  if (fstopaction=saHold) and (sHolding in st) then exit;

 // writeln(fStopChanged);

  // используется команда перемещения отн. позицию вместо команды остановки
  // это нужно для возможности сменить способ остановки, т.к. он игнорируется при команде остановки.
  // при этом перемещение в относит. позицию 0 косвенно дает остановку
  // однако если последний Stop не менял способ остановки - используем именно команду остановки
  if fStopChanged then begin
    ev.Attrs[ev3astop_action].sData:=ev3mStopActions[fstopaction];
    ev.Attrs[ev3aposition_sp].iData:=0;
    ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunToRelPos];
    fStopChanged:=false;
  end
  else
    ev.Attrs[ev3acommand].sData:=ev3mCommands[cStop];
end;


function TEv3TachoMotor.Run(sp: variant; sa: variant): TEv3TachoMotor;
begin
  setHowStop(sa);
  result:=self.Run(sp);
end;

function TEv3TachoMotor.Run(sp: variant): TEv3TachoMotor;

begin
 result:=self;
{
 setReverse(0);
 fspeed:=sp;
 ev.Attrs[ev3aduty_cycle_sp].sData:=sp;
 ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunDirect];
}
//setSpeed(sp);
fspeed:=sp;
setReverse(0);
//if sp<0 then setReverse(1) else setReverse(0);
ev.Attrs[ev3aspeed_sp].iData:=((fspeed*maxSpeed) div 100);
ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunForever];
end;

function TEv3TachoMotor.Run: TEv3TachoMotor;
begin
 result:=self;
 self.Run(fspeed);
end;

function TEv3TachoMotor.RunToAbsPos(pos: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
 setHowStop(sa);
 result:=self.runToAbsPos(pos,sp);
end;

function TEv3TachoMotor.RunToAbsPos(pos: variant; sp: variant): TEv3TachoMotor;
begin
 result:=self;
// setSpeed(sp);

 fspeed:=sp;
 setReverse(0);
 ev.Attrs[ev3aspeed_sp].iData:=((fspeed*maxSpeed) div 100);
// run-to-abs-pos: Runs the motor to an absolute position specified by``position_sp``
// and then stops the motor using the command specified in stop_action.

 ev.Attrs[ev3aposition_sp].sData:=pos;
 ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunToAbsPos];

end;

function TEv3TachoMotor.RunToAbsPos(pos: variant): TEv3TachoMotor;
begin
 result:=self.RunToAbsPos(pos,fspeed);
end;

function TEv3TachoMotor.RunTime(ms: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
  setHowStop(sa);
  result:=self.runTime(ms,sp);
end;

function TEv3TachoMotor.RunTime(ms: variant; sp: variant): TEv3TachoMotor;
begin
 result:=self;
// setSpeed(sp);

 fspeed:=sp;
 setReverse(0);
 ev.Attrs[ev3aspeed_sp].iData:=((fspeed*maxSpeed) div 100);

 ev.Attrs[ev3atime_sp].sData:=ms;
 ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunTimed];
end;

function TEv3TachoMotor.RunTime(ms: variant): TEv3TachoMotor;
begin
 result:=self.runtime(ms,fspeed);
end;

function TEv3TachoMotor.RunDeg(deg: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
 setHowStop(sa);
 result:=self.rundeg(deg,sp);
end;

function TEv3TachoMotor.RunDeg(deg: variant; sp: variant): TEv3TachoMotor;
begin
  result:=RunTurn(deg/360.0,sp);
end;

function TEv3TachoMotor.RunDeg(deg: variant): TEv3TachoMotor;
begin
  result:=RunTurn(deg/360.0);
end;

function TEv3TachoMotor.RunTurn(turn: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
 setHowStop(sa);
 result:=RunTurn(turn,sp);
end;

function TEv3TachoMotor.RunTurn(turn: variant; sp: variant): TEv3TachoMotor;
begin
 result:=self;
 if fcount_per_rot=0 then fcount_per_rot:=ev.Attrs[ev3acount_per_rot].iData;
 setSpeed(sp);
// writeln('turn - ',Round(double(turn)*fcount_per_rot));
 ev.Attrs[ev3aposition_sp].iData:=Round(double(turn)*fcount_per_rot);
 ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunToRelPos];
end;

function TEv3TachoMotor.RunTurn(turn: variant): TEv3TachoMotor;
begin
  result:=self;
  RunTurn(turn,fspeed);
end;

//--------------------------
procedure Done;
  var ev:TEv3Device;
      i,j:TEv3Port;
 begin
  for i:=outA to outD do
   for j:=outA to outD do begin
    if assigned(ev3Rules[i,j]) then freeAndNil(ev3Rules[i,j]);
    if assigned(ev3MRules[i,j]) then freeAndNil(ev3MRules[i,j]);
   end;

  for ev in ev3DeviceFactory do begin
    case ev.Device of
     ev3tLTachoMotor,ev3tMTachoMotor: begin
//      writeln('+++');
//      writeln(strerror(fpgeterrno));
//      ev.Attrs[ev3acommand].sData:=ev3mCommands[cReset];
//      writeln(ev3mCommands[cReset]);
//      ev.Attrs[ev3acommand].hardPutData(ev3mCommands[cReset]);
//      writeln(strerror(fpgeterrno));
      ev.Attrs[ev3astop_action].sData:=ev3mStopActions[saCoast];
      ev.Attrs[ev3acommand].sData:=ev3mCommands[cStop];
//      writeln('----');
    //  readln;
     end;
    end;
  end;
 end;

initialization
 pushDone(@Done);
end.

